import processing.serial.*;
Serial serialPort;

PFont font;

//  フィルタ
MovingAverageFilter filterX;
MovingAverageFilter filterY;
MovingAverageFilter filterZ;

//  オフセット
float  offsetX = 0;
float  offsetY = 0;
float  offsetZ = 0;
int  cab = 0;

//  ウインドウサイズ
int WIDTH = 640;
int HEIGHT = 480;

int v = 1000;

//  AMI602の測定データ格納領域
char[] data = new char[9];

//  角度
float x = 0;  //  roll
float y = 0;  //  pitch

//  円の中心位置
int OFFSET_X = WIDTH/2;
int OFFSET_Y = HEIGHT/2;

//  円の径
int CIRCLE_R = 200;

//  加速度センサの値
float xval = 0.0;
float yval = 0.0;
float zval = 0.0;

//  各チャンネルデータ
int [] ch = new int[6];

//  座標
float [] x1 = new float[36];
float [] y1 = new float[36];
float [] x2 = new float[36];
float [] y2 = new float[36];

int maxX = 0;
int maxY = 0;
int maxZ = 0;
int minX = 9999;
int minY = 9999;
int minZ = 9999;
int dir;

//  平均値フィルタ
public class MovingAverageFilter
{
    int taps;
    float values[];
    MovingAverageFilter(int requestedTaps) {
        taps = requestedTaps;
        values = new float[taps];
    }
    int processSample(int in) {
        float acc = 0;
        values[0] = in;
        for (int i = 0; i < taps; i++) {
            acc += values[i];
        }
        acc /= taps;
        for (int i = taps - 1; i > 0; i--) {
            values[i]   = values[i - 1];
        }
        return (int)acc;
    }
}

void setup() 
{
  size(WIDTH, HEIGHT);
  font = loadFont("CenturyGothic-Bold-36.vlw");
  textFont(font, 24);

  frameRate(30);
  
  filterX = new MovingAverageFilter(10);
  filterY = new MovingAverageFilter(10);
  filterZ = new MovingAverageFilter(10);
  //  環境に応じてシリアルポートを変更すること
  serialPort = new Serial(this,"COM17",9600);
  cab = 1;
  int idx;
  float radian;
  for (int i=0;i< 360; i+= 10) {
      idx = i / 10;
      radian = (float)i * PI / 180.0;

      x1[idx] = cos(radian) * CIRCLE_R - sin(radian) * 0;
      y1[idx] = cos(radian) * 0 - sin(radian) * CIRCLE_R;
      x1[idx] = x1[idx] + OFFSET_X;
      y1[idx] = y1[idx] + OFFSET_Y;

      x2[idx] = cos(radian) * (CIRCLE_R - 10) - sin(radian) * 0;
      y2[idx] = cos(radian) * 0 - sin(radian) * (CIRCLE_R - 10);
      x2[idx] = x2[idx] + OFFSET_X;
      y2[idx] = y2[idx] + OFFSET_Y;
  }
  
  
}
//  シリアルポートからのデータを待つ
void waitData()
{
    while (serialPort.available() <= 0) {
    }
}
//  データの受信と描画処理
void draw() 
{
  float radian, radian1, radian2;
  float x01,y01,x02,y02, x03,y03;
  float fx, fy, fz;
  int aveX = 0;
  int aveY = 0;
  int aveZ = 0;
  float xInDegree = 0.0;
  float yInDegree = 0.0;

  float ms = millis();
  //センサ値取得
  float rawX = 0;
  float rawY = 0;
  float rawZ = 0;
  //
  int dat;
  int cnt = 0;
  int teardown = 0;
  //
  int range0 = (maxX - minX);
  int range1 = (maxY - minY);
  int range2 = (maxZ - minZ);
  
  float v0 = 9999.0;
  float v1 = 9999.0;
  float v2 = 9999.0;
  float dirf = 0.0;
  float pitch, roll;
  float Xh, Yh,azimuth;
  //
  //  Arduinoからはシリアルポートを通して測定データが送信される
  //  測定データは"[XXXXXXXXX]"(前後に[]がつく)
  //  
  while (teardown != 1) {
    waitData();  //  データが届くのを待つ
    dat = serialPort.read();
    if (dat == '[') {  //  先頭データ？
      while (teardown != 1) {
        waitData();  //  データが届くのを待つ
        dat = serialPort.read();
        if (dat == ']' | cnt >= 9) {  //  データの最後？
          teardown = 1;  //  ループを抜ける
          break;
        }
        else {
          data[cnt] = (char)dat;  //  データをセット
        }
        cnt++;
      }
    }
  }
  //
  //  測定データ取り出し
  ch[0] = 0;
  ch[1] = 0;
  ch[2] = 0;
  ch[3] = 0;
  ch[4] = 0;
  ch[5] = 0;
  //  ch 1
  ch[0] = 0x0ff0 & (data[0] << 4);
  ch[0] = ch[0] | ((data[1] & 0xf0)>>4);  
  //  ch 2
  ch[1] = (0x0f & data[1]) << 8;
  ch[1] = ch[1] | data[2];  
  //  ch 3
  ch[2] = 0x0ff0 & (data[3] << 4);
  ch[2] = ch[2] | ((data[4] & 0xf0)>>4);  
  //  ch 4
  ch[3] = (0x0f & data[4]) << 8;
  ch[3] = ch[3] | data[5];  
  //  ch 5
  ch[4] = 0x0ff0 & (data[6] << 4);
  ch[4] = ch[4] | ((data[7] & 0xf0) >>4);  
  //  ch 6
  ch[5] = (data[7] & 0x0f) << 8;
  ch[5] = ch[5] | data[8];  
  
  rawX = (ch[4] - 2048) / 256;
  rawY = (ch[3] - 2048) / 256;
  rawZ = (ch[2] - 2048) / 256;

  //  何かキーが押されたときの値をオフセットとする
  if (cab == 1) {
    offsetX = rawX;
    offsetY = rawY;
    offsetZ = 0;
    cab = 0;
    minX = 9999;
    maxX = 0;
    minY = 9999;
    maxY = 0;
    minZ = 9999;
    maxZ = 0;
    
  }
  //
  //  地磁気の方角計算
  //
  //  最小値、最大値の保持
  if (maxX < ch[0]) {
    maxX = ch[0];
  }
  if (maxY < ch[1]) {
    maxY = ch[1];
  }
  if (maxZ < ch[5]) {
    maxZ = ch[5];
  }
  if (minX > ch[0]) {
    minX = ch[0];
  }
  if (minY > ch[1]) {
    minY = ch[1];
  }
  if (minZ > ch[5]) {
    minZ = ch[5];
  }
  
  if (range0 != 0) {
    v0 = (float)(ch[0] - minX) / (float)range0;
    v0 = v0 - 0.5;  //  -0.5 <-> 0.5
  }
  if (range1 != 0) {
    v1 = (float)(ch[1] - minY) / (float)range1;
    v1 = v1 - 0.5;  //  -0.5 <-> 0.5
  }
  if (range2 != 0) {
    v2 = (float)(ch[5] - minZ) / (float)range2;
    v2 = (v2 - 0.5); //  -0.5 <-> 0.5
   
  }
  //  オフセット分を引く
  rawX = rawX - offsetX;
  rawY = rawY - offsetY;
  rawZ = rawZ - offsetZ;
  
  aveX = filterX.processSample((int)(rawX * v));
  aveY = filterY.processSample((int)(rawY * v));
  aveZ = filterZ.processSample((int)(rawZ * v));
  
  fx = (float)aveX;
  fy = (float)aveY;
  fz = (float)aveZ;


  roll = x = atan2(aveX, aveZ);
  x = x / PI * 180;
  pitch = y = atan2(aveY, aveZ);
  y = y / PI * 180;
  //
  Xh = v1 * cos(pitch) + v2 * sin(roll) * sin(pitch) - v0 * cos(roll) * sin(pitch);  
  Yh = v2 * cos(roll) + v0 * sin(roll);
  azimuth  = atan2(Yh, Xh);
  dirf = (azimuth) / PI * 180;
  //

  //
  //  描画処理
  //  
  background(0);
  stroke(255, 255, 255);
  strokeWeight(4);
  fill(0);
  smooth();
  //  円
  arc(OFFSET_X,OFFSET_Y,CIRCLE_R * 2, CIRCLE_R * 2,0,TWO_PI);
  //  円の目盛
  for (int i=0;i< 360; i+= 10) {
      int idx = i / 10;
      line(x1[idx],y1[idx], x2[idx],y2[idx]);
      if (i == 60 || i == 120) {
        line(x1[idx],y1[idx], x1[idx], ((y1[idx] - OFFSET_Y) * -1) + OFFSET_Y);
      } 
  }
  // ロール
  radian = (float)x * PI / 180.0;
  //	X
  x01 = cos(radian) * CIRCLE_R;
  y01 = -1 * sin(radian) * CIRCLE_R;
  x01+=OFFSET_X;
  y01+=OFFSET_Y;

  x02 = cos(radian) * -1 * CIRCLE_R;
  y02 = sin(radian) * CIRCLE_R;
  x02+=OFFSET_X;
  y02+=OFFSET_Y;
  stroke(0, 255, 0);
  line(x01,y01, x02,y02);
  //  ピッチ
  radian = (float)y * PI / -180.0;
  //	Y
  x01 = cos(radian) * CIRCLE_R;
  y01 = -1 * sin(radian) * CIRCLE_R;
  x01+=OFFSET_X;
  y01+=OFFSET_Y;

  x02 = cos(radian) * -1 * CIRCLE_R;
  y02 = sin(radian) * CIRCLE_R;
  x02+=OFFSET_X;
  y02+=OFFSET_Y;
  stroke(0, 0, 255);
  line(x02, y01, x01, y01);
  //  方角
  dirf = dirf + 90.0;
  radian = (float)dirf * PI / -180.0;
  x01 = -1 * cos(radian) * (CIRCLE_R - 10);
  y01 = sin(radian) * (CIRCLE_R - 10);
  x01+=OFFSET_X;
  y01+=OFFSET_Y;
  
  x02 = cos(radian) * CIRCLE_R;
  y02 = -1 * sin(radian) * CIRCLE_R;
  x02+=OFFSET_X;
  y02+=OFFSET_Y;
  stroke(255, 0, 0);
  fill(255, 0 ,0);
  line(x02, y02, x01, y01);

  x01 = -1 * cos(radian) * (CIRCLE_R - 10);
  y01 = sin(radian) * (CIRCLE_R - 10);
  x01+=OFFSET_X;
  y01+=OFFSET_Y;
  radian1 = (float)(dirf + 5) * PI / -180.0;
  radian2 = (float)(dirf - 5) * PI / -180.0;
  x02 = -1 * cos(radian1) * (CIRCLE_R - 50);
  y02 = sin(radian1) * (CIRCLE_R - 50);
  x02+=OFFSET_X;
  y02+=OFFSET_Y;
  x03 = -1 * cos(radian2) * (CIRCLE_R - 50);
  y03 = sin(radian2) * (CIRCLE_R - 50);
  x03+=OFFSET_X;
  y03+=OFFSET_Y;
  triangle(x01,y01,x02,y02,x03,y03);
  
  //
  stroke(255, 255, 255);
  point(OFFSET_X, OFFSET_Y);

  //  文字の描画
  fill(255);
  
  textFont(font, 24);
  text("AMI602", 10, 30);
  
  text("time: " + ms / 1000, 510, 30);
  //センサ値 表示
  fill(255,255,255);
  text("CH4(X): " + ch[3], 5, 70); 
  text("CH5(Y): " + ch[4], 5, 100); 
  text("CH3(Z): " + ch[2], 5, 130);
  fill(255,255,0);
  text("CH1: " + ch[0], 520, 70);
  text("CH2: " + ch[1], 520, 100);
  text("CH6: " + ch[5], 520, 130); 
  
  fill(255,255,255);
  text("X: " + x, 5, 430);
  text("Y: " + y, 5, 460); 

  text("N", OFFSET_X - 10, OFFSET_Y - CIRCLE_R - 10); 
  text("S", OFFSET_X - 5, OFFSET_Y + CIRCLE_R + 30); 
  text("E", OFFSET_X + CIRCLE_R + 10, OFFSET_Y + 10); 
  text("W", OFFSET_X - CIRCLE_R - 30, OFFSET_Y + 10); 

}
//
void keyPressed() 
{
  cab = 1;
}
